from .query import Q
from .filter import F
from .aggs import A
from .function import SF
from .search import Search
from .field import *
from .document import DocType, MetaField
from .mapping import Mapping
from .index import Index
from .analysis import analyzer, token_filter, char_filter, tokenizer

VERSION = (0, 0, 4)
__version__ = VERSION
__versionstr__ = '.'.join(map(str, VERSION))
